import typing

import numpy as np
import torch
from jaxtyping import Float, Int
from muutils.json_serialize import serializable_dataclass, serializable_field
from torch.utils.data import Dataset

from maze_dataset import MazeDataset, MazeDatasetConfig
from maze_dataset.maze import PixelColors, SolvedMaze
from maze_dataset.maze.lattice_maze import PixelGrid, _remove_isolated_cells


def _extend_pixels(
    image: Int[np.ndarray, "x y rgb"], n_mult: int = 2, n_bdry: int = 1
) -> Int[np.ndarray, "n_mult*x+2*n_bdry n_mult*y+2*n_bdry rgb"]:
    wall_fill: int = PixelColors.WALL[0]
    assert all(
        x == wall_fill for x in PixelColors.WALL
    ), "PixelColors.WALL must be a single value"

    output: np.ndarray = np.repeat(
        np.repeat(
            image,
            n_mult,
            axis=0,
        ),
        n_mult,
        axis=1,
    )

    # pad on all sides by n_bdry
    output = np.pad(
        output,
        pad_width=((n_bdry, n_bdry), (n_bdry, n_bdry), (0, 0)),
        mode="constant",
        constant_values=wall_fill,
    )

    return output


_RASTERIZED_CFG_ADDED_PARAMS: list[str] = [
    "remove_isolated_cells",
    "extend_pixels",
    "endpoints_as_open",
]


def process_maze_rasterized_input_target(
    maze: SolvedMaze,
    remove_isolated_cells: bool = True,
    extend_pixels: bool = True,
    endpoints_as_open: bool = False,
) -> Float[torch.Tensor, "in/tgt=2 x y rgb=3"]:
    # problem and solution mazes
    maze_pixels: PixelGrid = maze.as_pixels(show_endpoints=True, show_solution=True)
    problem_maze: PixelGrid = maze_pixels.copy()
    solution_maze: PixelGrid = maze_pixels.copy()

    # in problem maze, set path to open
    problem_maze[(problem_maze == PixelColors.PATH).all(axis=-1)] = PixelColors.OPEN

    # wherever solution maze is PixelColors.OPEN, set it to PixelColors.WALL
    solution_maze[(solution_maze == PixelColors.OPEN).all(axis=-1)] = PixelColors.WALL
    # wherever it is solution, set it to PixelColors.OPEN
    solution_maze[(solution_maze == PixelColors.PATH).all(axis=-1)] = PixelColors.OPEN
    if endpoints_as_open:
        for color in (PixelColors.START, PixelColors.END):
            solution_maze[(solution_maze == color).all(axis=-1)] = PixelColors.OPEN

    # postprocess to match original easy_2_hard dataset
    if remove_isolated_cells:
        problem_maze = _remove_isolated_cells(problem_maze)
        solution_maze = _remove_isolated_cells(solution_maze)

    if extend_pixels:
        problem_maze = _extend_pixels(problem_maze)
        solution_maze = _extend_pixels(solution_maze)

    return torch.tensor(np.array([problem_maze, solution_maze]))


@serializable_dataclass
class RasterizedMazeDatasetConfig(MazeDatasetConfig):
    """
    - `remove_isolated_cells: bool` whether to set isolated cells to walls
    - `extend_pixels: bool` whether to extend pixels to match easy_2_hard dataset (2x2 cells, extra 1 pixel row of wall around maze)
    - `endpoints_as_open: bool` whether to set endpoints to open
    """

    remove_isolated_cells: bool = serializable_field(default=True)
    extend_pixels: bool = serializable_field(default=True)
    endpoints_as_open: bool = serializable_field(default=False)


class RasterizedMazeDataset(MazeDataset):
    def __getitem__(self, idx: int) -> Float[torch.Tensor, "item in/tgt=2 x y rgb=3"]:
        # get the solved maze
        solved_maze: SolvedMaze = self.mazes[idx]

        return process_maze_rasterized_input_target(
            maze=solved_maze,
            remove_isolated_cells=self.cfg.remove_isolated_cells,
            extend_pixels=self.cfg.extend_pixels,
            endpoints_as_open=self.cfg.endpoints_as_open,
        )

    def get_batch(
        self, idxs: list[int] | None
    ) -> Float[torch.Tensor, "in/tgt=2 item x y rgb=3"]:
        if idxs is None:
            idxs = list(range(len(self)))

        inputs: list[Float[torch.Tensor, "x y rgb=3"]]
        targets: list[Float[torch.Tensor, "x y rgb=3"]]
        inputs, targets = zip(*[self[i] for i in idxs])

        return torch.stack([torch.stack(inputs), torch.stack(targets)])

    @classmethod
    def from_config_augmented(
        cls,
        cfg: RasterizedMazeDatasetConfig,
        **kwargs,
    ) -> Dataset:
        """loads either a maze transformer dataset or an easy_2_hard dataset"""
        _cfg_temp: MazeDatasetConfig = MazeDatasetConfig.load(cfg.serialize())
        return cls.from_base_MazeDataset(
            cls.from_config(cfg=_cfg_temp, **kwargs),
            added_params={
                k: v
                for k, v in cfg.serialize().items()
                if k in _RASTERIZED_CFG_ADDED_PARAMS
            },
        )

    @classmethod
    def from_base_MazeDataset(
        cls,
        base_dataset: MazeDataset,
        added_params: dict | None = None,
    ) -> Dataset:
        """loads either a maze transformer dataset or an easy_2_hard dataset"""
        if added_params is None:
            added_params = dict(
                remove_isolated_cells=True,
                extend_pixels=True,
            )
        output: MazeDataset = cls(
            cfg=base_dataset.cfg,
            mazes=base_dataset.mazes,
        )
        cfg: RasterizedMazeDatasetConfig = RasterizedMazeDatasetConfig.load(
            {
                **base_dataset.cfg.serialize(),
                **added_params,
            }
        )
        output.cfg = cfg
        return output

    def plot(self, count: int | None = None, show: bool = True) -> tuple:
        import matplotlib.pyplot as plt

        print(f"{self[0][0].shape = }, {self[0][1].shape = }")
        count = count or len(self)
        if count == 0:
            print(f"No mazes to plot for dataset")
            return
        fig, axes = plt.subplots(2, count, figsize=(15, 5))
        if count == 1:
            axes = [axes]
        for i in range(count):
            axes[0, i].imshow(self[i][0])
            axes[1, i].imshow(self[i][1])
            # remove ticks
            axes[0, i].set_xticks([])
            axes[0, i].set_yticks([])
            axes[1, i].set_xticks([])
            axes[1, i].set_yticks([])

        if show:
            plt.show()

        return fig, axes


def make_numpy_collection(
    base_cfg: RasterizedMazeDatasetConfig,
    grid_sizes: list[int],
    from_config_kwargs: dict | None = None,
    verbose: bool = True,
    key_fmt: str = "{size}x{size}",
) -> dict[
    typing.Literal["configs", "arrays"],
    dict[str, RasterizedMazeDatasetConfig | np.ndarray],
]:
    """create a collection of configs and arrays for different grid sizes, in plain tensor form

    output is of structure:
    ```
    {
        "configs": {
            "<n>x<n>": RasterizedMazeDatasetConfig,
            ...
        },
        "arrays": {
            "<n>x<n>": np.ndarray,
            ...
        },
    }
    ```
    """

    if from_config_kwargs is None:
        from_config_kwargs = {}

    datasets: dict[int, RasterizedMazeDataset] = {}

    for size in grid_sizes:
        if verbose:
            print(f"Generating dataset for maze size {size}...")

        cfg_temp: RasterizedMazeDatasetConfig = RasterizedMazeDatasetConfig.load(
            base_cfg.serialize()
        )
        cfg_temp.grid_n = size

        datasets[size] = RasterizedMazeDataset.from_config_augmented(
            cfg=cfg_temp,
            **from_config_kwargs,
        )

    return dict(
        configs={
            key_fmt.format(size=size): dataset.cfg for size, dataset in datasets.items()
        },
        arrays={
            # get_batch(None) returns a single tensor of shape (n, 2, x, y, 3)
            key_fmt.format(size=size): dataset.get_batch(None).cpu().numpy()
            for size, dataset in datasets.items()
        },
    )



import functools
import json
import typing
import warnings
from pathlib import Path
from typing import Callable, Type

import numpy as np
import torch
from muutils.json_serialize import (
    JSONitem,
    SerializableDataclass,
    serializable_dataclass,
    serializable_field,
)
from muutils.misc import sanitize_fname, shorten_numerical_to_str, stable_hash
from muutils.mlutils import DEFAULT_SEED, GLOBAL_SEED, set_reproducibility
from muutils.tensor_utils import DTYPE_MAP
from torch.utils.data import Dataset
from zanj import ZANJ


class FilterInfoMismatchError(ValueError):
    """raised when the filter info in a dataset config does not match the filter info in the dataset"""

    pass


def _dtype_serialization_fn(datatype: torch.dtype | np.dtype) -> str:
    """convert torch dtype to string, while checking that the conversion is reversible"""
    x_str: str = str(datatype)
    assert x_str in DTYPE_MAP, f"unknown dtype {datatype}"
    assert DTYPE_MAP[x_str] == datatype
    return x_str


def _load_applied_filters(
    filters: list[dict[typing.Literal["name", "args", "kwargs"], str | list | dict]],
) -> list[dict[typing.Literal["name", "args", "kwargs"], str | list | dict]]:
    try:
        return [
            dict(
                name=filter_info["name"],
                args=tuple(
                    filter_info["args"]
                ),  # muutils/zanj save tuples as lists, and this causes problems
                kwargs=dict(filter_info["kwargs"]),
            )
            for filter_info in filters
        ]
    except Exception as e:
        raise ValueError(f"failed to load applied filters:\n{filters}") from e


@serializable_dataclass(kw_only=True)
class GPTDatasetConfig(SerializableDataclass):
    """base GPTDatasetConfig class"""

    name: str

    # TODO: get rid of all these things as part of migration to tokenizer-free dataset config
    # --------------------------------------------------
    seq_len_min: int = serializable_field(default=1)
    seq_len_max: int = serializable_field(default=512)
    # --------------------------------------------------

    seed: int | None = serializable_field(default=DEFAULT_SEED)
    applied_filters: list[
        dict[typing.Literal["name", "args", "kwargs"], str | list | dict]
    ] = serializable_field(
        default_factory=list,
        deserialize_fn=_load_applied_filters,
        assert_type=False,  # TODO: check the type here once muutils supports checking Callable signatures
    )

    def __post_init__(self):
        assert self.seq_len_min <= self.seq_len_max
        # if seed set to None, then generate a new random seed
        if self.seed is None:
            self.seed = torch.random.seed() % 2**31

        if (DEFAULT_SEED != self.seed) and (GLOBAL_SEED != self.seed):
            warnings.warn(
                f"in GPTDatasetConfig {self.name=}, {self.seed=} is trying to override {GLOBAL_SEED=} which has already been changed elsewhere from {DEFAULT_SEED=}"
            )

        set_reproducibility(self.seed)

    def summary(self) -> dict:
        """return a summary of the config"""
        self_ser: dict = self.serialize()
        return dict(
            name=self.name,
            seq_len_min=self.seq_len_min,
            seq_len_max=self.seq_len_max,
            seed=self.seed,
            applied_filters=self.applied_filters,
        )

    @classmethod
    @property
    def _dataset_class(cls) -> type:
        raise NotImplementedError("this should be implemented by subclasses!")

    def to_fname(self) -> str:
        """convert config to a filename"""
        self_json_str: str = json.dumps(self.serialize())
        self_json_hash: int = int(abs(stable_hash(self_json_str)) % 1e10)
        warnings.warn(
            f"using fallblack to_fname() method for {self.__class__.__name__}, this should be implemented by subclasses!"
        )
        return sanitize_fname(
            f"f{self.name}-n{shorten_numerical_to_str(len(self))}-h{self_json_hash}"
        )


def _dataset_config_load(*args, **kwargs) -> "GPTDatasetConfig":
    raise NotImplementedError(
        f"this `load` function should be implemented by subclasses! got: {args=}, {kwargs=}"
    )


def _dataset_config_serialize(self, *args, **kwargs) -> JSONitem:
    raise NotImplementedError(
        f"this `serialize` function should be implemented by subclasses! got: {args=}, {kwargs=}"
    )


GPTDatasetConfig.load = _dataset_config_load
GPTDatasetConfig.serialize = _dataset_config_serialize


class GPTDataset(Dataset):
    """wrapper for torch dataset with some extra functionality

    (meaning the functionality should be inherited in downstream classes)

    note: `GPTDatasetConfig` should implement a `to_fname` method that returns a unique filename for the config

    # Requires:
    the following methods should be implemented in subclasses:
        - `__init__(self, cfg: GPTDatasetConfig, **kwargs)`
        initialize the dataset from a given config. kwargs are not passed through, the kwargs should take the actual generated or loaded data (a list of objects or sequences probably)
        - `generate(cls, cfg: GPTDatasetConfig, **kwargs) -> GPTDataset`
        generate the dataset from a given config. kwargs are passed through from `from_config`, and should only contain things that dont belong in the config (i.e. how many threads to use for generation)
        - `serialize(self) -> JSONitem`
        serialize the dataset to a ZANJ-serializable object, including:
            - config
            - data in formats specified by `self.save_formats`
        - `load(cls, data: JSONitem) -> GPTDataset`
        load the dataset from a ZANJ-serializable object
        - `download(cls, cfg: GPTDatasetConfig, **kwargs) -> GPTDataset`
        given a config, try to download a dataset from some source. kwargs are passed through from `from_config`, and should only contain things that dont belong in the config (i.e. some kind of auth token or source url)
        - `__len__(self) -> int`
        return the length of the dataset, required for `torch.utils.data.Dataset`
        - `__getitem__(self, i: int) -> list[str]`
        return the ith item in the dataset, required for `torch.utils.data.Dataset`
        return the ith item in the dataset, required for `torch.utils.data.Dataset`
        -  `update_self_config(self) -> None`
        update the config of the dataset to match the current state of the dataset, used primarily in filtering and validation
        -  decorating the appropriate filter namespace with `register_filter_namespace_for_dataset(your_dataset_class)` if you want to use filters

    # Parameters:
        - `cfg : GPTDatasetConfig`
        config for the dataset, used to generate the dataset
        - `do_generate : bool`
        whether to generate the dataset if it isn't found
        (defaults to `True`)
        - `load_local : bool`
        whether to try finding the dataset locally
        (defaults to `True`)
        - `save_local : bool`
        whether to save the dataset locally if it is generated or downloaded
        (defaults to `True`)
        - `do_download : bool`
        whether to try downloading the dataset
        (defaults to `True`)
        - `local_base_path : Path`
        where to save the dataset
        (defaults to `Path("data/maze_dataset")`)

    # Returns:
        - `GPTDataset`
        the dataset, as you wanted it

    # Implements:
        - `save(self, file_path: str) -> None`
        save the dataset to a file, using ZANJ
        - `read(cls, file_path: str) -> GPTDataset`
        read the dataset from a file, using ZANJ
        get all items in the dataset, in the specified format
        - `filter_by(self)`
        returns a namespace class
        -  `_filter_namespace(self) -> Class`
        returns a namespace class for filtering the dataset, checking that method
        - `_apply_filters_from_config(self) -> None`
        apply filters to the dataset, as specified in the config. used in `from_config()` but only when generating

    """

    _FILTER_NAMESPACE: type = "this isn't a filter namespace! you have to initialize this by registering with `register_filter_namespace_for_dataset`"  # type: ignore

    @classmethod
    def from_config(
        cls,
        cfg: GPTDatasetConfig,
        do_generate: bool = True,
        load_local: bool = True,
        save_local: bool = True,
        zanj: ZANJ | None = None,
        do_download: bool = True,
        local_base_path: Path = Path("data/maze_dataset"),
        except_on_config_mismatch: bool = True,
        allow_generation_metadata_filter_mismatch: bool = True,
        verbose: bool = False,
        **kwargs,
    ) -> "GPTDataset":
        """base class for gpt datasets

        priority of loading:
        1. load from local
        2. download
        3. generate

        """

        print_log: Callable = print if verbose else lambda *_a, **_kw: None

        local_base_path = Path(local_base_path)
        fname: Path = Path(f"{cfg.to_fname()}.zanj")
        output: GPTDataset | None = None
        did_load_local: bool = False
        if zanj is None:
            zanj = ZANJ()

        print_log(f"trying to get the dataset '{cfg.to_fname()}'")

        if not (load_local or do_download or do_generate):
            raise ValueError(
                "no way to load dataset! you said not to load local, not to download, and not to generate"
            )

        dataset_path: Path = local_base_path / fname

        # try loading
        if load_local:
            if dataset_path.exists():
                print_log(f"loading dataset from {dataset_path.as_posix()}")
                try:
                    output = cls.read(dataset_path, zanj=zanj)
                    did_load_local = True
                    print_log("load successful!")
                except Exception as e:
                    print_log(f"failed to load dataset: {e}")

        if do_download and output is None:
            print_log("seeing if we can download the dataset...")
            try:
                output = cls.download(cfg, **kwargs)
                print_log("download successful!")
            except NotImplementedError:
                print_log("no download found, or download failed")

        if do_generate and output is None:
            print_log("generating dataset...")
            output = cls.generate(cfg, verbose=verbose, **kwargs)
            # only if we generated it, apply filters
            output = output._apply_filters_from_config()

        # check and save
        if output is None:
            raise ValueError("failed to load dataset!")

        cfg_diff: dict = cfg.diff(output.cfg, of_serialized=True)
        if cfg_diff:
            if except_on_config_mismatch:
                if allow_generation_metadata_filter_mismatch and (
                    cfg_diff
                    == {
                        "applied_filters": {
                            "self": [],
                            "other": [
                                {
                                    "name": "collect_generation_meta",
                                    "args": (),
                                    "kwargs": {},
                                }
                            ],
                        }
                    }
                ):
                    pass
                else:
                    raise ValueError(f"config mismatch: {cfg_diff = }")
            else:
                warnings.warn(f"config mismatch: {cfg_diff = }")

        if save_local and not did_load_local:
            print_log(f"saving dataset to {dataset_path}")
            output.save(dataset_path, zanj=zanj)

        print_log(
            f"Got dataset {output.cfg.name} with {len(output)} items. {output.cfg.to_fname() = }"
        )
        return output

    def save(self, file_path: Path | str, zanj: ZANJ | None = None):
        if zanj is None:
            zanj = ZANJ()
        zanj.save(self.serialize(), file_path)

    # serialization & loading
    @classmethod
    def read(cls, file_path: str, zanj: ZANJ | None = None) -> "GPTDataset":
        if zanj is None:
            zanj = ZANJ()
        return zanj.read(file_path)

    def serialize(self) -> JSONitem:
        raise NotImplementedError()

    def data_hash(self) -> int:
        raise NotImplementedError()

    @classmethod
    def load(cls, data: JSONitem) -> "GPTDataset":
        raise NotImplementedError()

    # generating & downloading
    @classmethod
    def generate(cls, cfg: GPTDatasetConfig, **kwargs) -> "GPTDataset":
        raise NotImplementedError()

    @classmethod
    def download(cls, cfg: GPTDatasetConfig, **kwargs) -> "GPTDataset":
        raise NotImplementedError()

    # filtering
    def update_self_config(self):
        """update the config of the dataset to match the actual data, if needed

        for example, adjust number of mazes after filtering
        """
        pass

    class FilterBy:
        """thanks GPT-4"""

        def __init__(self, dataset: "GPTDataset"):
            self.dataset: "GPTDataset" = dataset

        def __getattr__(self, name: str) -> typing.Callable[..., "GPTDataset"]:
            filter_func: DatasetFilterProtocol = getattr(
                self.dataset._FILTER_NAMESPACE, name
            )

            def wrapped_filter_func(*args, **kwargs):
                return filter_func(self.dataset, *args, **kwargs)

            return wrapped_filter_func

    @property
    def filter_by(self) -> "FilterBy":
        return self.FilterBy(self)

    def _apply_filters_from_config(self):
        """apply filters to the dataset, as specified in the config. used in `from_config()`"""
        output: GPTDataset = self
        # copy the list, and then clear it in the config. we do this because each time we apply a filter it will update config.applied_filters
        applied_filters_old: list[
            dict[typing.Literal["name", "args", "kwargs"], typing.Any]
        ] = self.cfg.applied_filters
        output.cfg.applied_filters = list()
        # apply the filters
        for filter_info in applied_filters_old:
            filter_name: str = filter_info["name"]
            if filter_name not in output._FILTER_NAMESPACE.__dict__:
                if filter_name.startswith("__custom__:"):
                    raise ValueError(
                        f"the dataset {output.cfg.to_fname()} was filtering using a custom filter: '{filter_name}', which we don't know about. add it to MazeDatasetFilters!"
                    )
                else:
                    raise ValueError(
                        f"the dataset {output.cfg.to_fname()} was filtering using an unknown filter: '{filter_name}'"
                    )
            filter_args: list = filter_info["args"] if "args" in filter_info else list()
            filter_kwargs: dict = (
                filter_info["kwargs"] if "kwargs" in filter_info else dict()
            )
            output = getattr(output.filter_by, filter_name)(
                *filter_args, **filter_kwargs
            )

        # update the config, perform checks
        # TODO: some funny business with manually specified filters here?
        output.update_self_config()
        _check_filter_equality(applied_filters_old, output.cfg.applied_filters)
        return output


def _check_filter_equality(
    filters_old: list[
        dict[typing.Literal["name", "args", "kwargs"], str | list | dict]
    ],
    filters_new: list[
        dict[typing.Literal["name", "args", "kwargs"], str | list | dict]
    ],
) -> None:
    try:
        assert len(filters_old) == len(filters_new)

        for filterinfo_new, filterinfo_old in zip(filters_old, filters_new):
            # basic checks
            assert isinstance(filterinfo_new, dict), "filterinfo_new is not a dict"
            assert isinstance(filterinfo_old, dict), "filterinfo_old is not a dict"
            assert all(
                key in filterinfo_new for key in ["name", "args", "kwargs"]
            ), "missing keys in filterinfo_new"
            assert all(
                key in filterinfo_old for key in ["name", "args", "kwargs"]
            ), "missing keys in filterinfo_old"

            # name
            assert (
                filterinfo_new["name"] == filterinfo_old["name"]
            ), "filter names don't match"

            # args
            assert len(filterinfo_new["args"]) == len(
                filterinfo_old["args"]
            ), "filter args of different lengths"
            for arg_new, arg_old in zip(filterinfo_new["args"], filterinfo_old["args"]):
                assert arg_new == arg_old, "filter args don't match"

            # kwargs
            assert len(filterinfo_new["kwargs"]) == len(
                filterinfo_old["kwargs"]
            ), "filter kwargs of different lengths"
            for key in filterinfo_old["kwargs"]:
                assert (
                    key in filterinfo_new["kwargs"]
                ), f"filter kwargs don't match: missing key '{key}'"
                assert (
                    filterinfo_new["kwargs"][key] == filterinfo_old["kwargs"][key]
                ), f"filter kwargs don't match: values for key '{key}' don't match"

    except AssertionError as e:
        raise FilterInfoMismatchError(
            f"config mismatch in applied filters: {filters_new} != {filters_old}"
        ) from e


def register_filter_namespace_for_dataset(
    dataset_cls: Type[GPTDataset],
) -> Callable[[Type], Type]:
    """register the namespace class with the given dataset class"""

    def decorator(filter_namespace_cls: Type) -> Type:
        dataset_cls._FILTER_NAMESPACE = filter_namespace_cls
        filter_namespace_cls._BASE_DATASET = dataset_cls

        return filter_namespace_cls

    return decorator


class DatasetFilterProtocol(typing.Protocol):
    def __call__(
        self,
        dataset: GPTDataset,
        **kwargs,
    ) -> GPTDataset: ...


def register_dataset_filter(
    method: DatasetFilterProtocol,
) -> DatasetFilterProtocol:
    """register a dataset filter, copying the underlying dataset and updating the config

    be sure to return a COPY, not the original?

    method should be a staticmethod of a namespace class registered with `register_filter_namespace_for_dataset`
    """

    @functools.wraps(method)
    def wrapper(dataset: GPTDataset, *args, **kwargs):
        new_dataset = method(dataset, *args, **kwargs)
        # update the config
        new_dataset.cfg.applied_filters.append(
            dict(name=method.__name__, args=args, kwargs=kwargs)
        )
        new_dataset.update_self_config()
        return new_dataset

    return wrapper



from typing import Any, Sequence

import matplotlib.pyplot as plt
import numpy as np


def plot_colored_text(
    tokens: Sequence[str],
    weights: Sequence[float],
    cmap: str | Any,  # assume its a colormap if not a string
    ax: plt.Axes = None,
    width_scale: float = 0.023,
    width_offset: float = 0.005,
    height_offset: float = 0.1,
    rect_height: float = 0.7,
    token_height: float = 0.7,
    label_height: float = 0.3,
    word_gap: float = 0.01,
    fontsize: int = 12,
    fig_height: float = 0.7,
    fig_width_scale: float = 0.25,
    char_min: int = 4,
):
    assert len(tokens) == len(
        weights
    ), f"The number of tokens and weights must be the same: {len(tokens)} != {len(weights)}"
    total_len_estimate: float = sum([max(len(tok), char_min) for tok in tokens])
    # set up figure if needed
    if ax is None:
        fig, ax = plt.subplots(
            figsize=(total_len_estimate * fig_width_scale, fig_height)
        )
    ax.axis("off")

    # Normalize the weights to be between 0 and 1
    norm_weights: Sequence[float] = (weights - np.min(weights)) / (
        np.max(weights) - np.min(weights)
    )

    # Create a colormap instance
    if isinstance(cmap, str):
        colormap = plt.get_cmap(cmap)
    else:
        colormap = cmap

    x_pos: float = 0.0
    for i, (tok, weight, norm_wgt) in enumerate(zip(tokens, weights, norm_weights)):
        color = colormap(norm_wgt)[:3]

        # Plot the background color
        rect_width = width_scale * max(len(tok), char_min)
        ax.add_patch(
            plt.Rectangle(
                (x_pos, height_offset),
                rect_width,
                height_offset + rect_height,
                fc=color,
                ec="none",
            )
        )

        # Plot the token
        ax.text(
            x_pos + width_offset,
            token_height,
            tok,
            fontsize=fontsize,
            va="center",
            ha="left",
        )

        # Plot the weight below the token
        ax.text(
            x_pos + width_offset,
            label_height,
            f"{weight:.2f}",
            fontsize=fontsize,
            va="center",
            ha="left",
        )

        x_pos += rect_width + word_gap

    return ax



import cProfile
import math
import pstats
import timeit
import typing
from typing import Any, Callable, Iterable, Literal, Mapping, NamedTuple, TypeVar

import numpy as np
from jaxtyping import Bool
from muutils.statcounter import StatCounter

WhenMissing = Literal["except", "skip", "include"]


def bool_array_from_string(
    string: str, shape: list[int], true_symbol: str = "T"
) -> Bool[np.ndarray, "*shape"]:
    """Transform a string into an ndarray of bools.

    Parameters
    ----------
    string: str
        The string representation of the array
    shape: list[int]
        The shape of the resulting array
    true_symbol:
        The character to parse as True. Whitespace will be removed. All other characters will be parsed as False.

    Returns
    -------
    np.ndarray
        A ndarray with dtype bool of shape `shape`

    Examples
    --------
    >>> bool_array_from_string(
    ...     "TT TF", shape=[2,2]
    ... )
    array([[ True,  True],
        [ True, False]])
    """
    stripped = "".join(string.split())

    expected_symbol_count = math.prod(shape)
    symbol_count = len(stripped)
    if len(stripped) != expected_symbol_count:
        raise ValueError(
            f"Connection List contains the wrong number of symbols. Expected {expected_symbol_count}. Found {symbol_count} in {stripped}."
        )

    bools = [True if symbol == true_symbol else False for symbol in stripped]
    return np.array(bools).reshape(*shape)


def corner_first_ndindex(n: int, ndim: int = 2) -> list[tuple]:
    """returns an array of indices, sorted by distance from the corner

    this gives the property that `np.ndindex((n,n))` is equal to
    the first n^2 elements of `np.ndindex((n+1, n+1))`

    ```
    >>> corner_first_ndindex(1)
    [(0, 0)]
    >>> corner_first_ndindex(2)
    [(0, 0), (0, 1), (1, 0), (1, 1)]
    >>> corner_first_ndindex(3)
    [(0, 0), (0, 1), (1, 0), (1, 1), (0, 2), (2, 0), (1, 2), (2, 1), (2, 2)]
    ```
    """

    unsorted: list = list(np.ndindex(tuple([n for _ in range(ndim)])))
    return sorted(unsorted, key=lambda x: (max(x), x if x[0] % 2 == 0 else x[::-1]))

    # alternate numpy version from GPT-4:
    """
    # Create all index combinations
    indices = np.indices([n]*ndim).reshape(ndim, -1).T
    # Find the max value for each index
    max_indices = np.max(indices, axis=1)
    # Identify the odd max values
    odd_mask = max_indices % 2 != 0
    # Make a copy of indices to avoid changing the original one
    indices_copy = indices.copy()
    # Reverse the order of the coordinates for indices with odd max value
    indices_copy[odd_mask] = indices_copy[odd_mask, ::-1]
    # Sort by max index value, then by coordinates
    sorted_order = np.lexsort((*indices_copy.T, max_indices))
    return indices[sorted_order]
    """


def adj_list_to_nested_set(adj_list: list) -> set:
    """Used for comparison of adj_lists

    Adj_list looks like [[[0, 1], [1, 1]], [[0, 0], [0, 1]], ...]
    We don't care about order of coordinate pairs within
    the adj_list or coordinates within each coordinate pair."""
    return {
        frozenset([tuple(start_coord), tuple(end_coord)])
        for start_coord, end_coord in adj_list
    }


_AM_K = typing.TypeVar("_AM_K")
_AM_V = typing.TypeVar("_AM_V")


def apply_mapping(
    mapping: Mapping[_AM_K, _AM_V],
    iter: Iterable[_AM_K],
    when_missing: WhenMissing = "skip",
) -> list[_AM_V]:
    """Given an and a mapping, apply the mapping to the iterable with certain options"""
    output: list[_AM_V] = list()
    item: _AM_K
    for item in iter:
        if item in mapping:
            output.append(mapping[item])
            continue
        match when_missing:
            case "skip":
                continue
            case "include":
                output.append(item)
            case "except":
                raise ValueError(f"item {item} is missing from mapping {mapping}")
            case _:
                raise ValueError(f"invalid value for {when_missing = }")
    return output


def apply_mapping_chain(
    mapping: Mapping[_AM_K, Iterable[_AM_V]],
    iter: Iterable[_AM_K],
    when_missing: WhenMissing = "skip",
) -> list[_AM_V]:
    """Given a list and a mapping, apply the mapping to the list"""
    output: list[_AM_V] = list()
    item: _AM_K
    for item in iter:
        if item in mapping:
            output.extend(mapping[item])
            continue
        match when_missing:
            case "skip":
                continue
            case "include":
                output.append(item)
            case "except":
                raise ValueError(f"item {item} is missing from mapping {mapping}")
            case _:
                raise ValueError(f"invalid value for {when_missing = }")
    return output


T = TypeVar("T")

FancyTimeitResult = NamedTuple(
    "FancyTimeitResult",
    [
        ("timings", StatCounter),
        ("return_value", T | None),
        ("profile", pstats.Stats | None),
    ],
)


def timeit_fancy(
    cmd: Callable[[], T] | str,
    setup: str = lambda: None,
    repeats: int = 5,
    namespace: dict[str, Any] | None = None,
    get_return: bool = True,
    do_profiling: bool = False,
) -> FancyTimeitResult:
    """
    Wrapper for `timeit` to get the fastest run of a callable with more customization options.

    Approximates the functionality of the %timeit magic or command line interface in a Python callable.

    # Parameters
    - `cmd: Callable[[], T] | str`
        The callable to time. If a string, it will be passed to `timeit.Timer` as the `stmt` argument.
    - `setup: str`
        The setup code to run before `cmd`. If a string, it will be passed to `timeit.Timer` as the `setup` argument.
    - `repeats: int`
        The number of times to run `cmd` to get a reliable measurement.
    - `namespace: dict[str, Any]`
        Passed to `timeit.Timer` constructor.
        If `cmd` or `setup` use local or global variables, they must be passed here. See `timeit` documentation for details.
    - `get_return: bool`
        Whether to pass the value returned from `cmd`. If True, the return value will be appended in a tuple with execution time.
        This is for speed and convenience so that `cmd` doesn't need to be run again in the calling scope if the return values are needed.
        (default: `False`)
    - `do_profiling: bool`
        Whether to return a `pstats.Stats` object in addition to the time and return value.
        (default: `False`)

    # Returns
    `FancyTimeitResult`, which is a NamedTuple with the following fields:
    - `time: float`
        The time in seconds it took to run `cmd` the minimum number of times to get a reliable measurement.
    - `return_value: T|None`
        The return value of `cmd` if `get_return` is `True`, otherwise `None`.
    - `profile: pstats.Stats|None`
        A `pstats.Stats` object if `do_profiling` is `True`, otherwise `None`.
    """
    timer: timeit.Timer = timeit.Timer(cmd, setup, globals=namespace)

    # Perform the timing
    times: list[float] = timer.repeat(repeats, 1)

    # Optionally capture the return value
    return_value: T | None = None
    profile: pstats.Stats | None = None

    if get_return or do_profiling:
        # Optionally perform profiling
        if do_profiling:
            profiler = cProfile.Profile()
            profiler.enable()

        return_value: T = cmd()

        if do_profiling:
            profiler.disable()
            profile = pstats.Stats(profiler).strip_dirs().sort_stats("cumulative")

    # reset the return value if it wasn't requested
    if not get_return:
        return_value = None

    return FancyTimeitResult(
        timings=StatCounter(times),
        return_value=return_value,
        profile=profile,
    )
